# coding: utf-8
#
# Copyright (c) 2023 Trevor Taylor
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# json schema represented as a python type var, e.g.:
#
# @dataclasses.dataclass
# class Address:
#  number: str|int
#  street: str
#
# Codec[dict[str,int|Address]].decode({'fred':3, 'jock':{'number':32, 'street': 'asler'}})
#
# For examples see json_codec.py.test
#

from xju.xn import Xn,in_context,in_function_context,readable_repr
from typing import TypeVar, Generic, Type, cast, Any, Protocol
from typing import _LiteralGenericAlias  # type: ignore
from types import GenericAlias, UnionType, NoneType
import xju.newtype

T=TypeVar('T')

JsonType = None|bool|dict|list|float|str


class Codec(Generic[T]):
    t:Type[T]
    def __init__(self, t: Type[T]):
        'initialse json decoder for type %(t)r'
        self.t=t
        self.codec:CodecProto=_explodeSchema(self.t)
        pass

    def __repr__(self):
        return f'{self.t!r} json codec'

    def encode(self,x:T) -> JsonType:
        'encode {self.t} {x} to json'
        return self.codec.encode(x)

    def decode(self,x:JsonType) -> T:
        'decode {x} to a {self.t}'
        try:
            return cast(T, self.codec.decode(x))
        except Exception:
            raise in_function_context(Codec.decode,vars()) from None
        pass

    def get_json_schema(self) -> dict:
        result = self.codec.get_json_schema()
        result.update({
            "$id": "https://example.com/address.schema.json",
            "$schema": "https://json-schema.org/draft/2020-12/schema"})
        return result
    pass

class CodecProto(Protocol):
    def encode(self, x:Any)->JsonType:
        pass
    def decode(self, x:JsonType)->Any:
        pass
    def get_json_schema(self) -> dict:
        pass
    pass

Atom=TypeVar('Atom',int,str,bool,float,None)

NewInt=TypeVar('NewInt',bound=xju.newtype.Int)
NewFloat=TypeVar('NewFloat',bound=xju.newtype.Float)
NewStr=TypeVar('NewStr',bound=xju.newtype.Str)

class NoopCodec(Generic[Atom]):
    t:Type[Atom]
    def __init__(self, t:Type[Atom]):
        self.t=t
    def encode(self, x:Atom)->Atom:
        if type(x) is not self.t and not (
                self.t is float and type(x) is int):
            raise Exception(f'{x!r} is not a {self.t}')
        return x
    def decode(self, x:Atom)->Atom:
        if type(x) is not self.t and not (
                self.t is float and type(x) is int):
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not a {self.t}')
        return x
    def get_json_schema(self) -> dict:
        st: str
        if self.t is int: st='integer'
        if self.t is str: st='string'
        if self.t is float: st='number'
        if self.t is bool: st='boolean'
        if self.t is None: st='null'
        return { 'type': st }
    pass

class NoneCodec:
    def encode(self, x:None)->None:
        if x is not None:
            raise Exception(f'{x!r} is not None')
        return x
    def decode(self, x:None)->None:
        if x is not None:
            raise Exception(f'{x!r} is not None')
        return x
    def get_json_schema(self) -> dict:
        return { 'type': 'null' }
    pass

class ListCodec:
    def __init__(self, value_codec:Codec):
        self.value_codec=value_codec
        pass
    def encode(self,x):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.encode(_) for _ in x]
    def decode(self,x):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.decode(_) for _ in x]
    def get_json_schema(self) -> dict:
        return {
            "type": "array",
            "items": self.value_codec.get_json_schema()
        }
    pass

class AnyListCodec:
    def __init__(self):
        self.value_codec=AnyJsonCodec()
        pass
    def encode(self,x):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.encode(_) for _ in x]
    def decode(self,x):
        if type(x) is not list:
            raise Exception(f'{x!r} is not a list')
        return [self.value_codec.decode(_) for _ in x]
    def get_json_schema(self) -> dict:
        return {
            "type": "array"
        }
    pass

class TupleCodec:
    def __init__(self,value_codecs):
        self.value_codecs=value_codecs
        self.number_of_codecs=len(value_codecs)
        pass
    def encode(self,x) -> list:
        'encode tuple {x} as a {self.number_of_codecs}-element list'
        try:
            if type(x) is not tuple:
                raise Exception(f'{x!r} is not a tuple')
            if len(x) != self.number_of_codecs:
                l=len(x)
                raise Exception(f'{x} does not have {self.number_of_codecs} items (it has {l} items)')
            return [c.encode(v) for c,v in zip(self.value_codecs,x)]
        except Exception:
            raise in_function_context(TupleCodec.encode,vars()) from None
        pass
    def decode(self,x) -> tuple:
        try:
            if type(x) is not list:
                raise Exception(f'{x!r} is not a list')
            if len(x) != self.number_of_codecs:
                l=len(x)
                raise Exception(f'{x} does not have {self.number_of_codecs} items (it has {l} items)')
            result=[c.decode(v) for c,v in zip(self.value_codecs,x)]
            return tuple(result)
        except Exception:
            raise in_function_context(TupleCodec.decode,vars()) from None
        pass
    def get_json_schema(self) -> dict:
        return {
            "type": "array",
            "prefixItems": [
                codec.get_json_schema() for codec in self.value_codecs
            ]
        }
    pass

class UnionCodec:
    def __init__(self,allowed_types):
        self.allowed_types=allowed_types
        self.value_codecs={t:_explodeSchema(t) for t in allowed_types}
        pass
    def encode(self,x) -> JsonType:
        'encode {x!r} as one of {self.allowed_types}'
        try:
            exceptions=[]
            for t, c in self.value_codecs.items():
                try:
                    return c.encode(x)
                except Exception as e:
                    exceptions.append( (t, e) )
                    pass
                pass
            raise Exception(' and '.join([f'failed to encode as {t} because {e}' for t,e in exceptions]))
                
        except Exception:
            raise in_function_context(UnionCodec.encode,vars()) from None
        pass
    def decode(self,x) -> tuple:
        try:
            exceptions=[]
            for t, c in self.value_codecs.items():
                try:
                    return c.decode(x)
                except Exception as e:
                    exceptions.append( (t, e) )
                    pass
                pass
            raise Exception(' and '.join([f'failed to decode as {t} because {e}' for t,e in exceptions]))
        except Exception:
            raise in_function_context(UnionCodec.decode,vars()) from None
        pass
    def get_json_schema(self) -> dict:
        return {
            "oneOf": [ codec.get_json_schema() for codec in self.value_codecs.values() ]
        }
    pass

class DictCodec:
    def __init__(self, key_codec, value_codec):
        self.key_codec=key_codec
        self.value_codec=value_codec
    def encode(self,x:dict)->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        result:dict={}
        for k,v in x.items():
            ek,ev=self.key_codec.encode(k),self.value_codec.encode(v)
            if not isinstance(ek, str):
                raise Exception(f'key encoder {self.key_codec} produced non-str {ek} from dict key {k}')
            result[ek]=ev
            pass
        return result
    def decode(self,x)->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        return {self.key_codec.decode(k):self.value_codec.decode(v) for k,v in x.items()}
    def get_json_schema(self) -> dict:
        return {
            'type': 'object',
            'additionalProperties': self.value_codec.get_json_schema()
        }
    pass

class AnyDictCodec:
    def __init__(self):
        self.key_codec=NoopCodec[str](str)
        self.value_codec=AnyJsonCodec()
    def encode(self,x:dict)->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        result:dict={}
        for k,v in x.items():
            ek,ev=self.key_codec.encode(k),self.value_codec.encode(v)
            result[ek]=ev
            pass
        return result
    def decode(self,x)->dict:
        if type(x) is not dict:
            raise Exception(f'{x!r} is not a dict')
        return {self.key_codec.decode(k):self.value_codec.decode(v) for k,v in x.items()}
    def get_json_schema(self) -> dict:
        return {
            'type': 'object'
        }
    pass

class AnyJsonCodec:
    def encode(self,x):
        # we assume x will be subsequently json.dumps()ed so defer validation to that
        return x
    def decode(self,x):
        # we assume x was json.loads()ed so assume it is appropriate
        return x
    def get_json_schema(self) -> dict:
        return {
            "oneOf": [ { 'type': t } for t in ['null','boolean','object','array','number','string'] ]
        }
    pass

class NewIntCodec(Generic[NewInt]):
    t:Type[NewInt]
    def __init__(self,t:Type[NewInt]):
        self.t=t
        self.base_codec=NoopCodec[int](int)
    def encode(self, x:NewInt)->int:
        if not isinstance(x, xju.newtype.Int):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value())
    def decode(self, x:JsonType)->NewInt:
        if not isinstance(x, int):
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not an int')
        return self.t(self.base_codec.decode(x))
    def get_json_schema(self) -> dict:
        return {
            'description': self.t.__name__,
            'type': 'integer'
        }
    pass

class NewFloatCodec(Generic[NewFloat]):
    t:Type[NewFloat]
    def __init__(self,t:Type[NewFloat]):
        self.t=t
        self.base_codec=NoopCodec[float](float)
    def encode(self, x:NewFloat)->float:
        if not isinstance(x, xju.newtype.Float):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value())
    def decode(self, x:JsonType)->NewFloat:
        if type(x) is not float and type(x) is not int:
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not a float (or an int)')
        return self.t(self.base_codec.decode(x))
    def get_json_schema(self) -> dict:
        return {
            'description': self.t.__name__,
            'type': 'number'
        }
    pass

class NewStrCodec(Generic[NewStr]):
    t:Type[NewStr]
    def __init__(self,t:Type[NewStr]):
        self.t=t
        self.base_codec=NoopCodec[str](str)
    def encode(self, x:NewStr)->str:
        if not isinstance(x, xju.newtype.Str):
            raise Exception(f'{x!r} is not a {self.t}')
        return self.base_codec.encode(x.value())
    def decode(self, x:JsonType)->NewStr:
        if type(x) is not str:
            t=type(x)
            raise Exception(f'{x!r} (of type {t}) is not an str')
        return self.t(self.base_codec.decode(x))
    def get_json_schema(self) -> dict:
        return {
            'description': self.t.__name__,
            'type': 'string'
        }
    pass

class LiteralStrCodec:
    value:str
    def __init__(self,value:str):
        self.value=value
    def encode(self, x:str)->str:
        if type(x) is not str or x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def decode(self, x:JsonType)->str:
        if type(x) is not str:
            raise Exception(f'{x!r} is not a string')
        if x != self.value:
            raise Exception(f'{x!r} is not {self.value!r}')
        return x
    def get_json_schema(self) -> dict:
        return {
            'type': 'string',
            'enum': [ self.value ]
        }
    pass

class ClassCodec:
    t:type
    attr_codecs:dict[str,Any]  # codec
    def __init__(self, t:type, attr_codecs:dict[str,Any]):
        self.t=t
        self.attr_codecs=attr_codecs
        pass
    def encode(self,x) -> dict:
        'encode {x} as a {self.t}'
        try:
            if type(x) is not self.t:
                xt=type(x)
                raise Exception(f'{x!r} (of type {xt}) is not a {self.t}')
            result={}
            for n, attr_codec in self.attr_codecs.items():
                try:
                    result[n]=attr_codec.encode(getattr(x,n))
                except Exception:
                    raise in_context(f'encode attribute {n}') from None
                pass
            return result
        except Exception:
            raise in_function_context(ClassCodec.encode,vars()) from None
        pass
    def decode(self,x) -> object:
        'deocde {x} as a {self.t}'
        try:
            attr_values=[]
            for n, attr_codec in self.attr_codecs.items():
                try:
                    if n not in x:
                        raise Exception(f'{x!r} has no {n!r} attribute')
                    value=attr_codec.decode(x[n])
                    if not isinstance(attr_codec,LiteralStrCodec):
                        attr_values.append(value)
                        pass
                except Exception:
                    raise in_context(f'decode attribute {n}') from None
                pass
            return self.t(*attr_values)
        except Exception:
            raise in_function_context(ClassCodec.decode,vars()) from None
        pass
    def get_json_schema(self) -> dict:
        return {
            'description': self.t.__name__,
            'type': 'object',
            'properties': {
                n: attr_codec.get_json_schema() for n, attr_codec in self.attr_codecs.items()
            }
        }
    pass
    
def _explodeSchema(t:type):
    '''explode type {t} into a tree of codecs'''
    try:
        if t is float:
            return NoopCodec[float](float)
        if t is int:
            return NoopCodec[int](int)
        if t is str:
            return NoopCodec[str](str)
        if t is bool:
            return NoopCodec[bool](bool)
        if t is None or t is NoneType:
            return NoneCodec()
        if t is list:
            return AnyListCodec()
        if type(t) is GenericAlias and getattr(t, '__origin__') is list:
            return ListCodec(_explodeSchema(getattr(t,'__args__')[0]))
        if type(t) is _LiteralGenericAlias:
            value = t.__args__[0]
            if not isinstance(value,str):
                raise Exception(f'{t!r} literal type is not supported (only string is implemented)')
            return LiteralStrCodec(value)
        if type(t) is GenericAlias and getattr(t, '__origin__') is tuple:
            return TupleCodec([_explodeSchema(_) for _ in getattr(t,'__args__')])
        if t is dict:
            return AnyDictCodec()
        if type(t) is GenericAlias and getattr(t, '__origin__') is dict:
            return DictCodec(*[_explodeSchema(_) for _ in getattr(t,'__args__')])
        if type(t) is UnionType:
            return UnionCodec(getattr(t,'__args__'))
        if issubclass(t,xju.newtype.Int):
            return NewIntCodec(t)
        if issubclass(t,xju.newtype.Float):
            return NewFloatCodec(t)
        if issubclass(t,xju.newtype.Str):
            return NewStrCodec(t)
        return ClassCodec(t, {n: _explodeSchema(nt) for n,nt in t.__annotations__.items()})
    except Exception:
        raise in_function_context(_explodeSchema,vars()) from None
    pass



# Copyright (c) 2022 Trevor Taylor
# coding: utf-8
# 
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that all
# copyright notices and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
# CMC - Context Management Classic i.e. non-async
#
from __future__ import annotations

import asyncio
import sys
import contextlib
from typing import TypeVar, Iterable, Dict as _Dict, overload, Tuple, Sequence, Union, Optional
from typing import ItemsView, KeysView
from typing import Mapping, Type, List, Generic, Any, Callable
from collections import OrderedDict
from collections.abc import Coroutine
import builtins
import threading

from xju.time import Duration
from xju.xn import in_function_context
from xju.assert_ import Assert

T = TypeVar('T', bound=contextlib.AbstractContextManager)

def cmclass(cls:Type[T]) -> Type[T]:
    '''Class decorator that adds context management __enter__ and __exit__
       that enter and exit all type-hinted attributes implementing contextlib.AbstractContextManager
       then any existing __enter__.
       Note that to satisfy mypy, the decorated class must already
       implement contextlib.AbstractContextManager and the preferred way to do that is
       to inherit from xju.cmc.CM - see cmc.py.test for examples.
    '''
    base_classes_to_enter = [ base_class for base_class in cls.__bases__
                              if issubclass(base_class, contextlib.AbstractContextManager) ]
    attrs_to_enter = [ n for n, t in cls.__annotations__.items()
                         if is_subclass(n, t, contextlib.AbstractContextManager)]
    # need a unique place to keep resources acquired by enter so exit can
    # exit them, only want resources for self, not subclasses (which do their own
    # handling). Note replacing . with _ could lead to clashes but there's no
    # way to avoid all clashes.
    resources_attr_name=f'{cls.__module__}.{cls.__name__}'.replace('.','_')
    orig_enter=cls.__dict__.get('__enter__',lambda self: self)
    def enter(self,
              base_classes_to_enter=base_classes_to_enter,
              attrs_to_enter=attrs_to_enter,
              resources_attr_name=resources_attr_name,
              orig_enter=orig_enter) -> Type[T]:
        with contextlib.ExitStack() as tentative:
            for base_class in base_classes_to_enter:
                if base_class is not CM:
                    cm = __ClassCm(base_class, self)
                    tentative.enter_context(cm)
                    pass
                pass
            for n in attrs_to_enter:
                tentative.enter_context(getattr(self, n))
                pass
            assert getattr(self,resources_attr_name, None) is None, f'{cls.__module__}.{cls.__name__} has a f{resources_attr_name} attribute already. Perhaps it is inherited multiple times (which is not allowed by xju.cmc.cmclass decorator)?'
            orig_enter(self)
            setattr(self,resources_attr_name,tentative.pop_all())
        return self
    orig_exit=cls.__dict__.get('__exit__',lambda self,t,e,b: None)
    def exit(self,t,e,b,resources_attr_name=resources_attr_name) -> None:
        resources = getattr(self,resources_attr_name)
        setattr(self,resources_attr_name,None)
        with contextlib.closing(resources):
            return orig_exit(self,t,e,b)
    old_sa=cls.__setattr__
    def sa(self, name, value, attrs_to_enter=attrs_to_enter):
        if getattr(self,resources_attr_name,None) and name in attrs_to_enter:
            assert False, f'xju.cmc {self} has been entered so cannot replace context-managed attribute {cls.__module__}.{cls.__name__}.{name}'
        return old_sa(self,name,value)
    setattr(cls, '__enter__', enter)
    setattr(cls, '__exit__', exit)
    setattr(cls, '__setattr__', sa)
    cls.__annotations__['__enter__'] = type(enter)
    cls.__annotations__['__exit__'] = type(exit)
    return cls

class CM(contextlib.AbstractContextManager):
    def __enter__(self):
        assert False, f'{CM.__module__}.CM.__enter__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        return self
    def __exit__(self, t, e, b) -> None:
        assert False, f'{CM.__module__}.CM.__exit__ not overridden, have you forgotten @cmclass on {self.__module__}.{self.__class__.__name__}?'
        pass

K = TypeVar('K')
V = TypeVar('V', bound=contextlib.AbstractContextManager)

class Dict(Mapping[K, V], contextlib.AbstractContextManager):
    '''dictionary that enters and exits its values, all of which must be context managers
       note that values may be added and removed before and/or after self is entered:
         before: the value will be entered when self is entered
         after:  the value will be entered before it is added (noting that any existing
                 value will be exited once (if) it is removed) - the enter-new-exit-old 
                 order might be important to you i.e. might not be what you need!'''
    entered = False
    x:_Dict[K,V]

    @overload
    def __init__(self):
        '''new empty Dict'''

    @overload
    def __init__(self, x:_Dict[K,V]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    @overload
    def __init__(self, x:Iterable[Tuple[K,V]]):
        '''initialise with value from {x} assuming those values have not been "entered"'''

    def __init__(self, *args, **kwargs):
        self.x = OrderedDict(*args, **kwargs)
        pass
    
    def __enter__(self):
        '''"enters" all values in order they were inserted'''
        with contextlib.ExitStack() as tentative:
            for k,v in self.x.items():
                tentative.enter_context(v)
            tentative.pop_all()
            pass
        self.entered = True
        return self

    def __exit__(self, t, e, b):
        '''"exits" all values in order they were inserted'''
        if self.entered:
            with contextlib.ExitStack() as resources:
                for k,v in self.x.items():
                    resources.push(v)
                    pass
                pass
            pass
        pass

    def __contains__(self, key:object) -> bool:
        return key in self.x

    def __delitem__(self, key:K):
        v = self.x[key]
        del self.x[key]
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        pass

    def __getitem__(self, key:K) -> V:
        return self.x[key]
    
    def __iter__(self):
        return iter(self.x.keys())

    def __len__(self) -> int:
        return len(self.x)

    def __repr__(self) -> str:
        return repr(self.x)

    def __setitem__(self, key:K, value:V) -> None:
        '''replace any current value of {key} with {value}'''
        '''- "enters" the new value and then "exits" any old value'''
        old = self.x.get(key,None)
        if old is not value:
            if self.entered:
                value.__enter__()
                if old is None:
                    self.x[key]=value
                else:
                    with contextlib.ExitStack() as f:
                        f.push(old)
                        self.x[key]=value
                    pass
                pass
            else:
                self.x[key]=value
            pass
        pass
    
    def __sizeof__(self) -> int:
        return self.x.__sizeof__()

    def clear(self):
        while len(self):
            self.popitem()
            pass
        pass

    def get(self, key, default=None):
        return self.x.get(key, default)

    def items(self) -> ItemsView[K,V]:
        '''items in order they were first inserted'''
        return self.x.items()

    def keys(self) -> KeysView[K]:
        '''keys in order they were first inserted'''
        return self.x.keys()

    @overload
    def pop(self, key:K) -> V:
        '''pop and "exit" value of {key} if self has a value for it
           otherwise raise KeyError'''

    @overload
    def pop(self, key:K, default:T) -> Union[V,T]:
        '''pop and "exit" value of {key} if self has a value for it
        otherwise return default'''

    def pop(self, key, default=None):
        if default is None:
            v = self.x.pop(key)
            if self.entered:
                with contextlib.ExitStack() as f:
                    f.push(v)
                pass
            return v
        else:
            try:
                v = self.x.pop(key)
                if self.entered:
                    with contextlib.ExitStack() as f:
                        f.push(v)
                    pass
                return v
            except KeyError:
                return default
            pass
        pass

    def popitem(self) -> Tuple[K, V]:
        '''pop and "exit" most recently added item'''
        k, v = self.x.popitem()
        if self.entered:
            with contextlib.ExitStack() as f:
                f.push(v)
                pass
            pass
        return k, v

    @overload
    def setdefault(self, key:K, default:V) -> V:
        '''set value of {key} to default "entered" if it has no value
           return value of {key}'''

    @overload
    def setdefault(self, key:K, default:None) -> Optional[V]:
        '''same as get(key,None)'''

    def setdefault(self, key, default):
        if default is not None:
            if not key in self:
                self[key] = default
            return self.x.get(key)
        else:
            return self.x.get(key, None)
        pass

    @overload
    def update(self, x:_Dict[K, V]):
        '''update self from {x} assuming its values are not yet entered'''

    @overload
    def update(self, x:Iterable[Tuple[K,V]]):
        '''update self from {x} assuming its values are not yet entered'''

    def update(self, x):
        if getattr(x,'keys',None):
            for k in x.keys():
                self[k] = x[k]
                pass
            pass
        else:
            for k,v in x:
                self[k] = v
                pass
            pass
        pass

class __ClassCm(contextlib.AbstractContextManager):
    '''target object {x} subclass {cls} context management methods
       - so for x: X where X(Y), __ClassCM(Y, x) will call Y.__enter__
         and Y.__exit__ directly avoiding any X.__enter__/__exit__
         overrides'''
    cls:Type
    x:object

    def __init__(self, cls, x):
        self.cls=cls
        self.x=x
        pass
    def __enter__(self):
        self.cls.__enter__(self.x)
        return self
    def __exit__(self, t, e, b):
        return self.cls.__exit__(self.x, t, e, b)
    pass

def is_subclass(n:str, t1, t2:type):
    '''check whether {n}'s type {t1} is a sub-class of {t2}'''
    try:
        Assert(t2).isInstanceOf(type)
        if isinstance(t1,type):
            return issubclass(t1, t2)
        else:
            return False
        pass
    except Exception:
        raise in_function_context(is_subclass,vars()) from None
    pass

ResultType=TypeVar('ResultType')
class Task(contextlib.AbstractAsyncContextManager,Generic[ResultType]):
    '''asyncio Task context manager that guarentees Task awaited before exit
       - note that normally, application will have waited for the task
         within the context; cancel-on-exit ensures clean up on exception
         e.g. when using gather() with its default return_exceptions=False'''
    coroutine: Coroutine[Any,Any,ResultType]
    task: None | asyncio.Task[ResultType]
    def __init__(self, coroutine: Coroutine[Any,Any,ResultType]):
        self.coroutine=coroutine
        self.task=None
    async def __aenter__(self) -> asyncio.Task[ResultType]:
        '''create and return task'''
        assert self.task is None
        self.task=asyncio.create_task(self.coroutine)
        return self.task
    async def __aexit__(self, t, e, b):
        '''cancel and await task'''
        assert self.task is not None
        task=self.task
        self.task=None
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

class Thread(Generic[ResultType],contextlib.AbstractContextManager[Callable[[],ResultType]]):
    __t: threading.Thread
    __result: ResultType | Exception
    
    def __init__(self, run: Callable[[],ResultType], stop: Callable[[],None]):
        '''at context entry call {run}() in new thread
           at context exit call {stop}() and wait for thread to exit
           result is not available until context exited'''
        self.run = run
        self.stop = stop
        self.__t = threading.Thread(target=self.__run)
        self.__result = Exception('thread not run')
        pass

    def __str__(self):
        return f'thread {self.__t} to run {self.run}'

    def __enter__(self) -> Callable[[],ResultType]:
        '''start {self}'''
        try:
            self.__t.start()
            return self.result
        except Exception:
            raise in_function_context(Thread.__enter__,vars())
        pass

    def __exit__(self, t, e, b):
        '''stop (via {self.stop}()) and wait for (i.e. join) {self}'''
        try:
            self.stop()
            self.__t.join()
        except Exception:
            raise in_function_context(Thread.__exit__, vars())
        pass

    def result(self) -> ResultType:
        '''assuming context has exited, return/raise result of run()'''
        if isinstance(self.__result, Exception):
            raise self.__result
        return self.__result

    def __run(self) -> None:
        try:
            self.__result = self.run()
        except Exception as e:
            self.__result = e
            pass
    pass

class Mutex:
    def __init__(self):
        self.m = threading.Lock()
        pass
    pass

class Lock(contextlib.AbstractContextManager):
    def __init__(self, m: Mutex):
        self.m = m
        self.active=False
        pass

    def __enter__(self):
        self.m.m.acquire()
        self.active=True
        return self

    def __exit__(self,t,e,b):
        self.active=False
        self.m.m.release()
        pass

    pass

class Condition:
    def __init__(self, m:Mutex):
        self.m = m
        self.c = threading.Condition(m.m)

    def wait_for(self, l:Lock, timeout:Duration):
        assert l.m == self.m and l.active
        self.c.wait(timeout=timeout.value())
        pass

    def notify_all(self, l:Lock):
        assert l.m == self.m and l.active
        self.c.notify_all()
    pass
